/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package nl.nmg.servlet;

import	java.io.BufferedReader;
import	java.io.File;
import	java.io.IOException;
import	java.sql.Date;
import	java.util.Enumeration;
import	java.util.Hashtable;
import	javax.servlet.ServletInputStream;
import	javax.servlet.http.Cookie;
import	javax.servlet.http.HttpServletRequest;
import	javax.servlet.http.HttpSession;

/**
 * ExtendedRequest is a wrapper object around an HttpServletRequest object, providing
 * multipart/form-data support and convenient getXXXParameter functions.
 *
 * <P>When an ExtendedRequest object is created, and the request content-type is
 * multipart/form-data, the data is parsed, and subsequent calls to
 * ExtendedRequest.getParameter() work as expected. Use getFileParameter() to
 * retrieve file upload data.
 *
 * <P>Additionally ExtendedRequest provides convenient retrieval functions like
 * getBooleanParameter, getIntParameter and getDateParameter.
 *
 * <P>All other methods just pass data on to the HttpServletRequest object.
 * This only works with JSDK2.0. You may need to add or remove a number
 * of dummy functions for other versions of the JSDK.
 */
public class ExtendedRequest implements HttpServletRequest {
	private HttpServletRequest	origReq;
	private Hashtable			params;

	/**
	 * Creates a ExtendedRequest object. If the input is of type multipart/form-data
	 * the input is parsed, and any uploaded file objects will be saved to
	 * a temporary directory. Remember to call <CODE>deleteTemporaryFiles()</CODE> when
	 * you are done, to delete these temporary files.
	 *
	 * @param request is the request object passed to the <CODE>service()</CODE>,
	 *		<CODE>doGet()</CODE> or <CODE>doPost()</CODE> method.
	 * @param tmpDir is the directory that may be used to hold temporary files.
	*/
	public ExtendedRequest(HttpServletRequest request, File tmpDir) throws IOException {
		origReq = request;
		if(request.getContentType() != null &&
				MultipartFormDataParser.getMIMEHeaderPrimaryValue(request.getContentType()).equals("multipart/form-data") &&
				request.getMethod() != null &&
				request.getMethod().equals("POST"))
		{
			MultipartFormDataParser	parser = new MultipartFormDataParser(request.getContentType(),
										request.getInputStream(), request.getContentLength(), tmpDir);
			params = parser.parse();
		} else {
			params = null;
		}
	}

	/**
	 * Creates a ExtendedRequest object. If the input is of type multipart/form-data
	 * the input is parsed, and any uploaded file objects will be saved in memory.
	 *
	 * @param request is the request object passed to the <CODE>service()</CODE>,
	 *		<CODE>doGet()</CODE> or <CODE>doPost()</CODE> method.
	*/
	public ExtendedRequest(HttpServletRequest request) throws IOException {
		this(request, null);
	}

	/**
	 * Deletes any temporary files created. This method should be called when
	 * the ExtendedRequest was created with the temporary directory option.
	 */
	public void deleteTemporaryFiles() {
		if(params != null)
			MultipartFormDataParser.deleteTemporaryFiles(params);
	}

	/**
	 * Returns the parameter names for this request as an enumeration
	 * of strings, or an empty enumeration if there are no parameters
	 * or the input stream is empty.
	 */
	public Enumeration getParameterNames() {
		if(params == null) {
			return origReq.getParameterNames();
		} else {
			return params.keys();
		}
	}

	/**
	 * Returns a string containing the lone value of the specified
	 * parameter, or null if the parameter does not exist. If the
	 * value is an uploaded file, the client-side filename
	 * is returned.
	 *
	 * <P>Servlet writers should use this method only when they are
	 * sure that
	 * there is only one value for the parameter.  If the parameter
	 * has (or could have) multiple values, servlet writers should
	 * use getParameterValues. If a multiple valued parameter name
	 * is passed as an argument, the return value is implementation
	 * dependent.
	 * </P>
	 *
	 * @param name the name of the parameter whose value is required.
	 * @see #getParameterValues
	 */
	public String getParameter(String name) {
		if(params == null) {
			return origReq.getParameter(name);
		} else {
			Object[]		vals;

			vals = (Object[]) params.get(name);
			if(vals == null) {
				return null;
			} else {
				return vals[0].toString();
			}
		}
	}

	/**
	 * Returns the values of the specified parameter for the request as
	 * an array of strings, or null if the named parameter does not
	 * exist. If any of the parameter's values are uploaded files, the
	 * client-side filenames are returned.
	 *
	 * @param name the name of the parameter whose value is required.
	 * @see #getParameter
	 */
	public String[] getParameterValues(String name) {
		if(params == null) {
			return origReq.getParameterValues(name);
		} else {
			Object[]		vals;
			String[]		stringVals;
			int				i;

			vals = (Object[]) params.get(name);
			if(vals == null) {
				return null;
			} else {
				stringVals = new String[vals.length];
				for(i = 0; i < vals.length; i++) {
					stringVals[i] = vals[i].toString();
				}
				return stringVals;
			}
		}
	}

	/**
	 * Returns the lone uploaded file value of the specified
	 * parameter, or null if the parameter does not exists
	 * or is not an uploaded file value.
	 *
	 * @param name the name of the parameter whose value is required.
	 * @see #getFileParameterValues
	 */
	public UploadedFile getFileParameter(String name) {
		if(params == null) {
			return null;
		} else {
			Object[]		vals;

			vals = (Object[]) params.get(name);
			if(vals == null) {
				return null;
			} else if(!(vals[0] instanceof UploadedFile)) {
				return null;
			} else {
				return (UploadedFile) vals[0];
			}
		}
	}

	/**
	 * Returns the uploaded file values of the specified parameter,
	 * or null if the specified parameter does not exist. The array
	 * contains null for a value that is not an uploaded file.
	 *
	 * @param name the name of the parameter whose value is required.
	 * @see #getFileParameter
	 */
	public UploadedFile[] getFileParameterValues(String name) {
		if(params == null) {
			return null;
		} else {
			Object[]		vals;
			UploadedFile[]	fileVals;
			int				i;

			vals = (Object[]) params.get(name);
			if(vals == null) {
				return null;
			} else {
				fileVals = new UploadedFile[vals.length];
				for(i = 0; i < vals.length; i++) {
					if(vals[i] instanceof UploadedFile) {
						fileVals[i] = (UploadedFile) vals[i];
					} else {
						fileVals[i] = null;
					}
				}
				return fileVals;
			}
		}
	}

	/**
	 * Returns true when the specified parameter exists, false
	 * otherwise.
	 *
	 * @param name the name of the parameter whose value is required.
	 */
	public boolean getBooleanParameter(String name) {
		return getParameter(name) != null;
	}

	/**
	 * Returns the lone value of the specified parameter converted to
	 * an <CODE>int</CODE>, or the specified default if the value can not be
	 * converted or the parameter does not exist.
	 *
	 * @param name the name of the parameter whose value is required
	 * @param defaultValue the default value to use
	 */
	public int getIntParameter(String name, int defaultValue) {
		String val = getParameter(name);

		if(val != null) {
			try {
				return Integer.parseInt(val);
			} catch(NumberFormatException nfe) {
				//
			}
		}
		return defaultValue;
	}

	/**
	 * Returns the lone value of the specified parameter converted to
	 * a <CODE>double</CODE>, or the specified default if the value can not be
	 * converted or the parameter does not exist.
	 *
	 * @param name the name of the parameter whose value is required
	 * @param defaultValue the default value to use
	 */
	public double getDoubleParameter(String name, double defaultValue) {
		String val = getParameter(name);

		if(val != null) {
			try {
				return Double.valueOf(val).doubleValue();
			} catch(NumberFormatException nfe) {
				//
			}
		}
		return defaultValue;
	}

	/**
	 * Returns the date value associated with the three parameters
	 * that start with the specified prefix. The prefix is appended
	 * with "<CODE>_day</CODE>", "<CODE>_month</CODE>" and "<CODE>_year</CODE>"
	 * to find the day of the month, the month and the year respectivily.
	 * 
	 * <P>The day is a number between 1 and 31 (inclusive) and the month is
	 * a number between 1 and 12 (inclusive). The year is the real year,
	 * <EM>not</EM> the year minus 1900. If the conversion failed or one of
	 * the parameters does not exist, <CODE>null</CODE> is returned.
	 *
	 * @param prefix the prefix of the parameter whose value is required
	 * @see nl.nmg.servlet.HTMLWriter#printDateSelect
	 */
	public Date getDateParameter(String prefix) {
		String	dayVal, monthVal, yearVal;
		int		day, month, year;

		dayVal = getParameter(prefix + "_day");
		monthVal = getParameter(prefix + "_month");
		yearVal = getParameter(prefix + "_year");
		if(dayVal != null && dayVal.length() > 0 &&
			monthVal != null && monthVal.length() > 0 &&
			yearVal != null && yearVal.length() > 0)
		{
			try {
				return new Date(
					Integer.parseInt(yearVal) - 1900,
					Integer.parseInt(monthVal) - 1,
					Integer.parseInt(dayVal));
			} catch(NumberFormatException nfe) {
				//
			}
		}
		return null;
	}


	/*
	 * These methods implement the ServletRequest interface
	 */

	public Object getAttribute(String name) {
		return origReq.getAttribute(name);
	}

	public String getCharacterEncoding() {
		return origReq.getCharacterEncoding();
	}


	public int getContentLength() {
		return origReq.getContentLength();
	}

	public String getContentType() {
		return origReq.getContentType();
	}

	public ServletInputStream getInputStream() throws IOException {
		return origReq.getInputStream();
	}

	public String getProtocol() {
		return origReq.getProtocol();
	}

	public BufferedReader getReader() throws IOException {
		return origReq.getReader();
	}

	public String getRealPath(String virtualPath) {
		return origReq.getRealPath(virtualPath);
	}

	public String getRemoteAddr() {
		return origReq.getRemoteAddr();
	}

	public String getRemoteHost() {
		return origReq.getRemoteHost();
	}

	public String getScheme() {
		return origReq.getScheme();
	}

	public String getServerName() {
		return origReq.getServerName();
	}

	public int getServerPort() {
		return origReq.getServerPort();
	}

	/*
	 * These methods implement the HttpServletRequest interface
	 */

	public String getAuthType() {
		return origReq.getAuthType();
	}

	public Cookie[] getCookies() {
		return origReq.getCookies();
	}

	public long getDateHeader(String name) {
		return origReq.getDateHeader(name);
	}

	public String getHeader(String name) {
		return origReq.getHeader(name);
	}

	public Enumeration getHeaderNames() {
		return origReq.getHeaderNames();
	}

	public int getIntHeader(String name) {
		return origReq.getIntHeader(name);
	}

	public String getMethod() {
		return origReq.getMethod();
	}

	public String getPathInfo() {
		return origReq.getPathInfo();
	}

	public String getPathTranslated() {
		return origReq.getPathTranslated();
	}

	public String getQueryString() {
		return origReq.getQueryString();
	}

	public String getRemoteUser() {
		return origReq.getRemoteUser();
	}

	public String getRequestedSessionId() {
		return origReq.getRequestedSessionId();
	}

	public String getRequestURI() {
		return origReq.getRequestURI();
	}

	public String getServletPath() {
		return origReq.getServletPath();
	}

	public HttpSession getSession(boolean create) {
		return origReq.getSession(create);
	}

	public boolean isRequestedSessionIdFromCookie() {
		return origReq.isRequestedSessionIdFromCookie();
	}

	public boolean isRequestedSessionIdFromUrl() {
		return origReq.isRequestedSessionIdFromUrl();
	}

	public boolean isRequestedSessionIdValid() {
		return origReq.isRequestedSessionIdValid();
	}
}
