/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	nl.nmg.servlet;
import	java.io.BufferedReader;
import	java.io.IOException;
import	java.io.OutputStream;
import	java.io.PrintWriter;
import	java.io.StringReader;
import	java.io.Writer;
import	java.text.SimpleDateFormat;
import	java.util.Calendar;
import	java.util.Date;

/**
 * The HTMLWriter class is a subclass for a <CODE>PrintWriter</CODE>
 * that adds methods for printing HTML elements, especially form elements.
 */
public class HTMLWriter extends PrintWriter {
	private static final String[]	monthNames =
		{"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

	/**
	 * Create a new HTMLWriter, without automatic line flushing.
	 *
	 * @param  out		  A character-output stream
	 */
	public HTMLWriter(Writer out) {
		super(out);
	}

	/**
	 * Create a new HTMLWriter.
	 *
	 * @param  out		  A character-output stream
	 * @param  autoFlush  A boolean; if true, the println() methods will flush
	 *					  the output buffer
	 */
	public HTMLWriter(Writer out, boolean autoFlush) {
		super(out, autoFlush);
	}

	/**
	 * Create a new HTMLWriter, without automatic line flushing, from an
	 * existing OutputStream.  This convenience constructor creates the
	 * necessary intermediate OutputStreamWriter, which will convert characters
	 * into bytes using the default character encoding.
	 *
	 * @param  out		  An output stream
	 *
	 * @see java.io.OutputStreamWriter
	 */
	public HTMLWriter(OutputStream out) {
		super(out);
	}

	/**
	 * Create a new HTMLWriter from an existing OutputStream.  This
	 * convenience constructor creates the necessary intermediate
	 * OutputStreamWriter, which will convert characters into bytes using the
	 * default character encoding.
	 *
	 * @param  out		  An output stream
	 * @param  autoFlush  A boolean; if true, the println() methods will flush
	 *					  the output buffer
	 *
	 * @see java.io.OutputStreamWriter
	 */
	public HTMLWriter(OutputStream out, boolean autoFlush) {
		super(out, autoFlush);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="hidden"&gt;</CODE> form element.
	 * If the value is <CODE>null</CODE>, a comment is printed instead.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 * @param value the value of the <CODE>value</CODE> attribute
	 *			or <CODE>null</CODE> if no form element should be printed.
	 */
	public void printHiddenInput(String name, String value) {
		if(value != null) {
			print("<INPUT type=hidden name=\"");
			print(HTMLEncoder.encode(name));
			print("\" value=\"");
			print(HTMLEncoder.encode(value));
			println("\">");
		} else {
			println("<!-- " + name + " is null -->");
		}
	}



	private void printSimpleInput(String type, String name, String value, int size, int maxlength) {
		print("<INPUT type=");
		print(type);
		if(name != null) {
			print(" name=\"");
			print(HTMLEncoder.encode(name));
			print("\"");
		}
		if(value != null) {
			print(" value=\"");
			print(HTMLEncoder.encode(value));
			print("\"");
		}
		if(size != -1) {
			print(" size=");
			print(size);
		}
		if(maxlength != -1) {
			print(" maxlength=");
			print(maxlength);
		}
		println(">");
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="text"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param maxlength the value of the <CODE>maxlength</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printTextInput(String name, String value, int size, int maxlength) {
		printSimpleInput("text", name, value, size, maxlength);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="text"&gt;</CODE> form element
	 * without a <CODE>maxlength</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printTextInput(String name, String value, int size) {
		printSimpleInput("text", name, value, size, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="text"&gt;</CODE> form element
	 * without <CODE>size</CODE> and <CODE>maxlength</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printTextInput(String name, String value) {
		printSimpleInput("text", name, value, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="text"&gt;</CODE> form element without
	 * <CODE>value</CODE>, <CODE>size</CODE> and <CODE>maxlength</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printTextInput(String name) {
		printSimpleInput("text", name, null, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="password"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param maxlength the value of the <CODE>maxlength</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printPasswordInput(String name, String value, int size, int maxlength) {
		printSimpleInput("password", name, value, size, maxlength);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="password"&gt;</CODE> form element
	 * without a <CODE>maxlength</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printPasswordInput(String name, String value, int size) {
		printSimpleInput("password", name, value, size, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="password"&gt;</CODE> form element
	 * without <CODE>size</CODE> and <CODE>maxlength</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printPasswordInput(String name, String value) {
		printSimpleInput("password", name, value, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="password"&gt;</CODE> form element without
	 * <CODE>value</CODE>, <CODE>size</CODE> and <CODE>maxlength</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printPasswordInput(String name) {
		printSimpleInput("password", name, null, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="file"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param maxlength the value of the <CODE>maxlength</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printFileInput(String name, int size, int maxlength) {
		printSimpleInput("file", name, null, size, maxlength);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="file"&gt;</CODE> form element
	 * without a <CODE>maxlength</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printFileInput(String name, int size) {
		printSimpleInput("file", name, null, size, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="file"&gt;</CODE> form element
	 * without <CODE>size</CODE> and <CODE>maxlength</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printFileInput(String name) {
		printSimpleInput("file", name, null, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="submit"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printSubmitInput(String name, String value) {
		printSimpleInput("submit", name, value, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="submit"&gt;</CODE> form element
	 * without a <CODE>value</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printSubmitInput(String value) {
		printSimpleInput("submit", null, value, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="submit"&gt;</CODE> form element
	 * without <CODE>name</CODE> and <CODE>value</CODE> attributes.
	 */
	public void printSubmitInput() {
		printSimpleInput("submit", null, null, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="reset"&gt;</CODE> form element.
	 *
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printResetInput(String value) {
		printSimpleInput("reset", null, value, -1, -1);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="reset"&gt;</CODE> form element
	 * without a <CODE>value</CODE> attribute.
	 */
	public void printResetInput() {
		printSimpleInput("reset", null, null, -1, -1);
	}



	private void printBooleanInput(String type, String name, String value, boolean checked, String label) {
		print("<INPUT type=");
		print(type);
		if(name != null) {
			print(" name=\"");
			print(HTMLEncoder.encode(name));
			print("\"");
		}
		if(value != null) {
			print(" value=\"");
			print(HTMLEncoder.encode(value));
			print("\"");
		}
		if(checked) {
			print(" checked");
		}
		if(label != null) {
			print("> ");
			print(HTMLEncoder.encode(label));
			println();
		} else {
			println(">");
		}
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 * @param label the label to follow the checkbox
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printCheckboxInput(String name, String value, boolean checked, String label) {
		printBooleanInput("checkbox", name, value, checked, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element
	 * without a label.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 */
	public void printCheckboxInput(String name, String value, boolean checked) {
		printCheckboxInput(name, value, checked, null);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element
	 * without a <CODE>value</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 * @param label the label to follow the checkbox
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printCheckboxInput(String name, boolean checked, String label) {
		printCheckboxInput(name, null, checked, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element
	 * without a label or a <CODE>value</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 */
	public void printCheckboxInput(String name, boolean checked) {
		printCheckboxInput(name, null, checked, null);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element
	 * without <CODE>value</CODE> and <CODE>checked</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param label the label to follow the checkbox
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printCheckboxInput(String name, String label) {
		printCheckboxInput(name, null, false, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="checkbox"&gt;</CODE> form element
	 * without a label or <CODE>value</CODE> and <CODE>checked</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printCheckboxInput(String name) {
		printCheckboxInput(name, null, false, null);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 * @param label the label to follow the radio
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printRadioInput(String name, String value, boolean checked, String label) {
		printBooleanInput("radio", name, value, checked, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element
	 * without a label.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value of the <CODE>value</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 */
	public void printRadioInput(String name, String value, boolean checked) {
		printRadioInput(name, value, checked, null);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element
	 * without a <CODE>value</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 * @param label the label to follow the radio
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printRadioInput(String name, boolean checked, String label) {
		printRadioInput(name, null, checked, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element
	 * without a label or a <CODE>value</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param checked the value of the <CODE>checked</CODE> attribute
	 */
	public void printRadioInput(String name, boolean checked) {
		printRadioInput(name, null, checked, null);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element
	 * without <CODE>value</CODE> and <CODE>checked</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param label the label to follow the radio
	 *				or <CODE>null</CODE> if there should be no label.
	 */
	public void printRadioInput(String name, String label) {
		printRadioInput(name, null, false, label);
	}

	/**
	 * Prints a <CODE>&lt;INPUT type="radio"&gt;</CODE> form element
	 * without a label or <CODE>value</CODE> and <CODE>checked</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printRadioInput(String name) {
		printRadioInput(name, null, false, null);
	}



	/**
	 * Prints a <CODE>&lt;TEXTAREA&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param cols the value of the <CODE>cols</CODE> attribute
	 *				or -1 if it should not be set
	 * @param rows the value of the <CODE>rows</CODE> attribute
	 *				or -1 if it should not be set
	 * @param wrap the value of the <CODE>wrap</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param value the value between the <CODE>&lt;TEXTAREA&gt;</CODE> and
	 *				<CODE>&lt;/TEXTAREA&gt;</CODE> tags or <CODE>null</CODE>
	 *				if it should be empty.
	 */
	public void printTextArea(String name, int cols, int rows, String wrap, String value) {
		print("<TEXTAREA name=\"");
		print(HTMLEncoder.encode(name));
		print("\"");
		if(cols != -1) {
			print(" cols=");
			print(cols);
		}
		if(rows != -1) {
			print(" rows=");
			print(rows);
		}
		if(wrap != null) {
			print(" wrap=\"");
			print(HTMLEncoder.encode(wrap));
			print("\"");
		}
		print(">");
		if(value != null) {
			print(HTMLEncoder.encode(value));	// FIXME: shouldn't we encode newlines as real newlines?
		}
		println("</TEXTAREA>");
	}

	/**
	 * Prints an empty <CODE>&lt;TEXTAREA&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param cols the value of the <CODE>cols</CODE> attribute
	 *				or -1 if it should not be set
	 * @param rows the value of the <CODE>rows</CODE> attribute
	 *				or -1 if it should not be set
	 * @param wrap the value of the <CODE>wrap</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printTextArea(String name, int cols, int rows, String wrap) {
		printTextArea(name, cols, rows, wrap, null);
	}

	/**
	 * Prints an empty <CODE>&lt;TEXTAREA&gt;</CODE> form element
	 * without a <CODE>wrap</CODE> attribute.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param cols the value of the <CODE>cols</CODE> attribute
	 *				or -1 if it should not be set
	 * @param rows the value of the <CODE>rows</CODE> attribute
	 *				or -1 if it should not be set
	 */
	public void printTextArea(String name, int cols, int rows) {
		printTextArea(name, cols, rows, null, null);
	}

	/**
	 * Prints an empty <CODE>&lt;TEXTAREA&gt;</CODE> form element
	 * without <CODE>wrap</CODE>, <CODE>cols</CODE> and
	 * <CODE>rows</CODE> attributes.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 */
	public void printTextArea(String name) {
		printTextArea(name, -1, -1, null, null);
	}



	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param values the values for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 * @param selected the value of the <CODE>&lt;OPTION&gt;</CODE> that should
	 *				have a <CODE>selected</CODE> attribute.
	 * @param labels the labels for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 */
	public void printSingleSelect(String name, String[] values, String selected, String[] labels) {
		print("<SELECT");
		if(name != null) {
			print(" name=\"");
			print(HTMLEncoder.encode(name));
			print("\"");
		}
		println(">");

		for(int i = 0; i < values.length; i++) {
			print("\t<OPTION");
			if(values[i] != null) {
				print(" value=\"");
				print(HTMLEncoder.encode(values[i]));
				print("\"");
			}
			if(selected != null && values[i].equals(selected)) {
				print(" selected");
				selected = null;
			}
			print("> ");
			if(labels[i] != null) {
				println(HTMLEncoder.encode(labels[i]));
			} else if(values[i] != null) {
				println(HTMLEncoder.encode(values[i]));
			} else {
				println();
			}
		}

		println("</SELECT>");
	}

	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element with the values
	 * for labels.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param values the values and labels for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 * @param selected the value of the <CODE>&lt;OPTION&gt;</CODE> that should
	 *				have a <CODE>selected</CODE> attribute.
	 */
	public void printSingleSelect(String name, String[] values, String selected) {
		printSingleSelect(name, values, selected, values);
	}

	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element with the values
	 * for labels, without a selected option.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param values the values and labels for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 */
	public void printSingleSelect(String name, String[] values) {
		printSingleSelect(name, values, null, values);
	}



	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element for multiple selections.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param values the values for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 * @param selected an array of booleans of the same size as <CODE>values</CODE>.
	 *				If one of the booleans is true, the corresponding option
	 *				gets a <CODE>selected</CODE> attribute.
	 * @param labels the labels for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 */
	public void printMultipleSelect(String name, int size, String[] values, boolean[] selected, String[] labels) {
		print("<SELECT");
		if(name != null) {
			print(" name=\"");
			print(HTMLEncoder.encode(name));
			print("\"");
		}
		if(size != -1) {
			print(" size=");
			print(size);
		}
		println(" multiple>");

		for(int i = 0; i < values.length; i++) {
			print("\t<OPTION");
			if(values[i] != null) {
				print(" value=\"");
				print(HTMLEncoder.encode(values[i]));
				print("\"");
			}
			if(selected[i]) {
				print(" selected");
			}
			print("> ");
			if(labels[i] != null) {
				println(HTMLEncoder.encode(labels[i]));
			} else if(values[i] != null) {
				println(HTMLEncoder.encode(values[i]));
			} else {
				println();
			}
		}

		println("</SELECT>");
	}

	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element for multiple selections
	 * with the values for labels.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param values the values for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 * @param selected an array of booleans of the same size as <CODE>values</CODE>.
	 *				If one of the booleans is true, the corresponding option
	 *				gets a <CODE>selected</CODE> attribute.
	 */
	public void printMultipleSelect(String name, int size, String[] values, boolean[] selected) {
		printMultipleSelect(name, size, values, selected, values);
	}

	/**
	 * Prints a <CODE>&lt;SELECT&gt;</CODE> form element for multiple selections
	 * with the values for labels, without a selected option.
	 *
	 * @param name the value of the <CODE>name</CODE> attribute
	 *				or <CODE>null</CODE> if it should not be set
	 * @param size the value of the <CODE>size</CODE> attribute
	 *				or -1 if it should not be set
	 * @param values the values for the <CODE>&lt;OPTION&gt;</CODE> form elements.
	 */
	public void printMultipleSelect(String name, int size, String[] values) {
		printMultipleSelect(name, size, values, new boolean[values.length], values);
	}


	/**
	 * Prints three <CODE>&lt;SELECT&gt;</CODE> form elements that
	 * can be used to select a date. "<CODE>_day</CODE>",
	 * "<CODE>_month</CODE>" and "<CODE>_year</CODE>" are added to the
	 * prefix to make the name attributes for the form elements.
	 *
	 * @param prefix the prefix for the <CODE>name</CODE> attributes
	 * @param lowestYear the lowest year to display in the year selection element
	 * @param highestYear the highest year to display in the year selection element
	 * @param selected the date that should be selected in the selection element
	 *				or <CODE>null</CODE> if no date should be selected
	 * @param allowNull is true when the form should allow for empty dates
	 * @see nl.nmg.servlet.ExtendedRequest#getDateParameter
	 */
	public void printDateSelect(String prefix, int lowestYear, int highestYear, Date selected, boolean allowNull) {
		Calendar	cal = null;
		int			i;

		if(selected != null) {
			cal = Calendar.getInstance();
			cal.setTime(selected);
		}

		// print day select
		print("<SELECT name=\"");
		print(prefix);
		println("_day\">");

		if(allowNull) {
			print("\t<OPTION value=\"\"");
			if(cal == null) {
				print(" selected");
			}
			println("> DD");
		}
		for(i = 1; i <= 31; i++) {
			print("\t<OPTION value=");
			print(i);
			if(cal != null && cal.get(Calendar.DAY_OF_MONTH) == i) {
				print(" selected> ");
			} else {
				print("> ");
			}
			print(i);
			println();
		}

		println("</SELECT>");

		// print month select
		print("<SELECT name=\"");
		print(prefix);
		println("_month\">");

		if(allowNull) {
			print("\t<OPTION value=\"\"");
			if(cal == null) {
				print(" selected");
			}
			println("> MM");
		}
		for(i = 1; i <= 12; i++) {
			print("\t<OPTION value=");
			print(i);
			if(cal != null && (cal.get(Calendar.MONTH)+1) == i) {
				print(" selected> ");
			} else {
				print("> ");
			}
			print(monthNames[i-1]);
			println();
		}

		println("</SELECT>");

		// print year select
		print("<SELECT name=\"");
		print(prefix);
		println("_year\">");

		if(allowNull) {
			print("\t<OPTION value=\"\"");
			if(cal == null) {
				print(" selected");
			}
			println("> YYYY");
		}
		for(i = lowestYear; i <= highestYear; i++) {
			print("\t<OPTION value=");
			print(i);
			if(cal != null && cal.get(Calendar.YEAR) == i) {
				print(" selected> ");
			} else {
				print("> ");
			}
			print(i);
			println();
		}

		println("</SELECT>");
	}

	/**
	 * Prints three <CODE>&lt;SELECT&gt;</CODE> form elements that
	 * can be used to select a date. "<CODE>_day</CODE>",
	 * "<CODE>_month</CODE>" and "<CODE>_year</CODE>" are added to the
	 * prefix to make the name attributes for the form elements. The
	 * form elements do not allow for empty dates.
	 *
	 * @param prefix the prefix for the <CODE>name</CODE> attributes
	 * @param lowestYear the lowest year to display in the year selection element
	 * @param highestYear the highest year to display in the year selection element
	 * @param selected the date that should be selected in the selection element
	 *				or <CODE>null</CODE> if no date should be selected.
	 * @see nl.nmg.servlet.ExtendedRequest#getDateParameter
	 */
	public void printDateSelect(String prefix, int lowestYear, int highestYear, Date selected) {
		printDateSelect(prefix, lowestYear, highestYear, selected, true);
	}

	/**
	 * Prints three <CODE>&lt;SELECT&gt;</CODE> form elements that
	 * can be used to select a date. "<CODE>_day</CODE>",
	 * "<CODE>_month</CODE>" and "<CODE>_year</CODE>" are added to the
	 * prefix to make the name attributes for the form elements. The
	 * form elements do not allow for empty dates, no date is selected.
	 *
	 * @param prefix the prefix for the <CODE>name</CODE> attributes
	 * @param lowestYear the lowest year to display in the year selection element
	 * @param highestYear the highest year to display in the year selection element
	 * @see nl.nmg.servlet.ExtendedRequest#getDateParameter
	 */
	public void printDateSelect(String prefix, int lowestYear, int highestYear) {
		printDateSelect(prefix, lowestYear, highestYear, null, true);
	}



	/**
	 * Prints free-form ascii text as nicely formatted HTML code.
	 * Depending upon the value of <CODE>allowHtml</CODE>
	 * the following transformations are done:
	 * <UL>
	 * <LI>&lt; and &gt; are encoded as &amp;lt; and &amp;gt; when they
	 *		do not appear to belong to a HTML tag, or when <CODE>allowHtml</CODE>
	 *		is false.
	 * <LI>&amp; is encoded as &amp;amp; when it does not appear to belong
	 *		to an HTML entity, or when <CODE>allowHtml</CODE> is false.
	 * <LI>URL's and email addresses that do not appear to be inside a HTML tag,
	 *		are translated surrounded by appropiate <CODE>&lt;A HREF&gt;</CODE>
	 *		tags.
	 * <LI>asterisks that appear to form a bullet list are replaced by
	 *		a sequence of <CODE>&lt;UL&gt;</CODE>, <CODE>&lt;LI&gt;</CODE>
	 *		and <CODE>&lt;/UL&gt;</CODE> tags.
	 * <LI>newlines that appear to be there to break the flow of the text
	 *		are prefixed with <CODE>&lt;BR&gt;</CODE>.
	 * </UL>
	 *
	 * <P>This method is meant to be used to print text entered by people
	 * that know little to no HTML. It tries to do its best to generate
	 * nice looking HTML, but the algorithm has to guess a lot of the time.
	 * So do not be surprised when the output is not what you expected.
	 * Have a look at the code, and try to fix it. :)
	 *
	 * @param in a <CODE>BufferedReader</CODE> that can be read to get the text
	 * @param allowHtml is true when HTML tags are allowed (you may
	 *			want to set this to <CODE>false</CODE> when making
	 *			something like a guestbook)
	 * @exception java.io.IOException when an I/O error occurs
	 */
	public void printText(BufferedReader in, boolean allowHtml) throws IOException {
		String			s, line;
		StringBuffer	translatedLine;
		char			c, d;
		int				i, j, k;
		boolean			startedTag = false, inBulletList = false;

		translatedLine = new StringBuffer();
		while((line = in.readLine()) != null) {
			translatedLine.setLength(0);
			i = 0;
			LINESCAN: while(i < line.length()) {
				c = line.charAt(i);

				switch(c) {
			// encode possibly unwanted characters
				case '<':
					if(allowHtml &&
							i < line.length() - 1 &&
							(Character.isLetter(line.charAt(i+1)) || line.charAt(i+1) == '/'))
					{
						translatedLine.append(c);
						startedTag = true;
					} else {
						translatedLine.append("&lt;");
					}
					i++;
					continue LINESCAN;
				case '>':
					if(startedTag) {
						translatedLine.append(c);
						startedTag = false;
					} else {
						translatedLine.append("&gt;");
					}
					i++;
					continue LINESCAN;
				case '&':
					if(allowHtml &&
							i < line.length() - 3 &&
							(Character.isLetter(line.charAt(i+1)) || line.charAt(i+1) == '#') &&
							(j = line.indexOf(';', i)) != -1 &&
							j < i + 6)
					{
						translatedLine.append(c);
					} else {
						translatedLine.append("&amp;");
					}
					i++;
					continue LINESCAN;
			// automatically link URLs
				case 'f':
				case 'F':
				case 'h':
				case 'H':
				case 'm':
				case 'M':
					// enough room for a valid URL and not already within a tag?
					if(i < line.length()-14 && !startedTag) {
						d = line.charAt(i+1);
						// is next char a T or an A?
						if(d == 't' || d == 'T' || d == 'a' || d == 'A') {
							// does URL start with 'http://', 'ftp://', 'https://' or 'mailto:'?
							if(line.substring(i, i + 7).equalsIgnoreCase("http://") ||
								line.substring(i, i + 6).equalsIgnoreCase("ftp://") ||
								line.substring(i, i + 8).equalsIgnoreCase("https://") ||
								line.substring(i, i + 7).equalsIgnoreCase("mailto:"))
							{
								// URL not part of a HTML tag parameter?
								if(i < 2 || line.charAt(i-1) != '\"' || line.charAt(i-2) != '=') {
									// gather characters for URL until an invalid character is encountered
									for(j = i + 6; j < line.length(); j++) {
										d = line.charAt(j);
										if(Character.isWhitespace(d) ||
												d == '<' || d == '>' || d == '&' ||
												d == '\"' || d == '\'' || d == ',')
											break;
									}
									// is URL long enough?
									if(j > i + 14) {
										// if URL ends with a period, remove it
										d = line.charAt(j-1);
										if(d == '.' || d == '!' || d == '?')
											j--;
										// we now have the complete URL and we will 
										s = line.substring(i , j);
										translatedLine.append("<A HREF=\"").append(s).append("\">").append(s).append("</A>");
										i = j;
										continue LINESCAN;
									}
								}
							}
						}
					}
					break;
			// automatically link email addresses
				case '@':
					// enough room for a valid email address and not already within a tag?
					if(i > 3 && i < line.length()-5 && !startedTag) {
						// gather characters before @
						for(j = i-1; j >= 0; j--) {
							d = line.charAt(j);
							if(!Character.isLetterOrDigit(d) && d != '.' && d != '-' && d != '_' && d != '%') {
								j++;
								break;
							}
						}
						if(j < 0)
							j = 0;

						// gather characters after @
						for(k = i+1; k < line.length(); k++) {
							d = line.charAt(k);
							if(!Character.isLetterOrDigit(d) && d != '.' && d != '-') {
								break;
							}
						}

						// is email address long enough?
						if(i-j > 3 && k-i > 5) {
							translatedLine.setLength(translatedLine.length() - (i-j));
							s = line.substring(j, k);
							translatedLine.append("<A HREF=\"mailto:").append(s).append("\">").append(s).append("</A>");
							i = k;
							continue LINESCAN;
						}
					}
					break;
				}
				translatedLine.append(c);
				i++;
			}

			// check for bullet list
			if(line.length() > 0 && line.charAt(0) == '*') {
				// make bullet list
				if(!inBulletList) {
					println("<UL>");
					inBulletList = true;
				}
				// print <LI>
				translatedLine.setCharAt(0, ' ');
				print("<LI>");
			} else if(inBulletList &&
					(line.length() == 0 || !Character.isWhitespace(line.charAt(0))))
			{
				println("</UL>");
				inBulletList = false;
			}
	
			// print generated line
			print(translatedLine.toString());

			// print break if necessary
			if(!allowHtml || (
				line.indexOf("<BR>") == -1 &&
				line.indexOf("<P>") == -1 &&
				line.indexOf("</P>") == -1 &&
				line.indexOf("<HR>") == -1 &&
				line.indexOf("<TABLE") == -1 &&
				line.indexOf("</TABLE") == -1 &&
				line.indexOf("<TR") == -1 &&
				line.indexOf("</TR") == -1 &&
				line.indexOf("<TH") == -1 &&
				line.indexOf("</TH") == -1 &&
				line.indexOf("<TD") == -1 &&
				line.indexOf("</TD") == -1))
			{
				println("<BR>");
			} else {
				println();
			}
		}

		if(inBulletList) {
			println("</UL>");
		}
	}

	/**
	 * Prints free-form ascii text as nicely formatted HTML code.
	 * HTML tags are allowed.
	 *
	 * @param in a <CODE>BufferedReader</CODE> that can be read to get the text
	 * @exception java.io.IOException when an I/O error occurs
	 */
	public void printText(BufferedReader in) throws IOException {
		printText(in, true);
	}

	/**
	 * Prints free-form ascii text as nicely formatted HTML code.
	 *
	 * @param text the text
	 * @param allowHtml is true when HTML tags are allowed (you may
	 *			want to set this to <CODE>false</CODE> when making
	 *			something like a guestbook)
	 */
	public void printText(String text, boolean allowHtml) {
		try {
			printText(new BufferedReader(new StringReader(text)), allowHtml);
		} catch(IOException ioexc) {
			throw new IllegalStateException("Internal error: we got an I/O Exception on a StringReader");
		}
	}

	/**
	 * Prints free-form ascii text as nicely formatted HTML code.
	 * HTML tags are allowed.
	 *
	 * @param text the text
	 */
	public void printText(String text) {
		printText(text, true);
	}



	/**
	 * Prints a date (day, month and year) using a SimpleDateFormat.
	 *
	 * @param format the format to use (see the documentation for SimpleDateFormat).
	 * @param value the date to print
	 * @see java.text.SimpleDateFormat
	 */
	public void printDate(String format, Date value) {
		SimpleDateFormat	sdf;

		sdf = new SimpleDateFormat(format);
		print(sdf.format(value));
	}

	/**
	 * Prints a date (day, month and year) using SimpleDateFormat's default format.
	 *
	 * @param value the date to print
	 * @see java.text.SimpleDateFormat
	 */
	public void printDate(Date value) {
		SimpleDateFormat	sdf;

		sdf = new SimpleDateFormat();
		print(sdf.format(value));
	}
}
