/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	nl.nmg.servlet;

import	java.io.InputStream;
import	java.io.IOException;
import	java.io.OutputStream;
import	java.sql.Connection;
import	java.sql.PreparedStatement;
import	java.sql.ResultSet;
import	java.sql.SQLException;
import	javax.servlet.http.HttpServletResponse;

/**
 * The UploadedFileUtils class contains a number of utility methods
 * to handle the moving of uploaded files to and from a database.
 *
 * <P><EM>Sample code to demonstrate this should be provided.</EM>
 */
public class UploadedFileUtils {
	/**
	 * No public constructor
	 */
	private UploadedFileUtils() {
	}

	/**
	 * Displays a file input form element, and optionally a file-delete checkbox.
	 *
	 * @param rs a result set set at the location of the row that contains the file.
	 * @param sqlColumn the column in the table that contains the file data. If the
	 *		column name is "<CODE>image</CODE>", the column "<CODE>image_ct</CODE>"
	 *		should contain the content-type and the column "<CODE>image_fn</CODE>"
	 *		should contain the client-side filename.
	 * @param out the writer where the form element should be printed.
	 * @exception java.sql.SQLException when a database related error occurs
	 */
	public static void editFile(ResultSet rs, String sqlColumn, HTMLWriter out) throws SQLException {
		String		val;

		out.printFileInput(sqlColumn);

		val = rs.getString(sqlColumn + "_fn");
		if(val != null) {
			out.println("<BR>");
			out.print("<INPUT type=checkbox name=\"");
			out.print(HTMLEncoder.encode(sqlColumn));
			out.print("_delete\"> Remove current file (");
			out.print(HTMLEncoder.encode(val));
			out.println(")");
		}
	}

	/**
	 * Copies the contents and meta-data of the specified file into the database.
	 * If the file-delete checkbox was checked, the data is set to <CODE>NULL</CODE>
	 * instead.
	 *
	 * @param extReq the request object that contains the file
	 * @param conn the database connection
	 * @param sqlTable the table in the database that contains the file
	 * @param sqlColumn the column in the table that contains the file data. If the
	 *		column name is "<CODE>image</CODE>", the column "<CODE>image_ct</CODE>"
	 *		should contain the content-type and the column "<CODE>image_fn</CODE>"
	 *		should contain the client-side filename.
	 * @param sqlCondition a WHERE clause that selects the appropiate row
	 * @exception java.sql.SQLException when a database related error occurs
	 */
	public static void updateFile(ExtendedRequest extReq, Connection conn,
							String sqlTable, String sqlColumn, String sqlCondition) throws SQLException, IOException
	{
		UploadedFile		uplFile;
		PreparedStatement	stmt;

		uplFile = extReq.getFileParameter(sqlColumn);
		if(uplFile != null && uplFile.getContentFullFilename() != null &&
			uplFile.getContentFullFilename().length() > 0)
		{
			stmt = conn.prepareStatement("UPDATE " + sqlTable + " SET " + sqlColumn + " = ?, " +
										sqlColumn + "_fn = ?, " + sqlColumn + "_ct = ? " + sqlCondition);
			try {
				stmt.setEscapeProcessing(false);
				stmt.setBinaryStream(1, uplFile.getInputStream(), uplFile.getContentLength());
				stmt.setString(2, uplFile.getContentFilename());
				stmt.setString(3, uplFile.getContentType());
				stmt.executeUpdate();
			} finally {
				stmt.close();
			}
		} else if(extReq.getBooleanParameter(sqlColumn + "_delete")) {
			stmt = conn.prepareStatement("UPDATE " + sqlTable + " SET " + sqlColumn + " = NULL, " +
										sqlColumn + "_fn = NULL, " + sqlColumn + "_ct = NULL " + sqlCondition);
			try {
				stmt.setEscapeProcessing(false);
				stmt.executeUpdate();
			} finally {
				stmt.close();
			}	
		}
	}

	/**
	 * Writes the contents of the specified file to the output stream of
	 * the HttpServletResponse object. The content-type is also set correctly.
	 *
	 * @param conn the database connection
	 * @param sqlTable the table in the database that contains the file
	 * @param sqlColumn the column in the table that contains the file data. If the
	 *		column name is "<CODE>image</CODE>", the column "<CODE>image_ct</CODE>"
	 *		should contain the content-type and the column "<CODE>image_fn</CODE>"
	 *		should contain the client-side filename.
	 * @param sqlCondition a WHERE clause that selects the appropiate row
	 * @param res the target of the file
	 * @exception java.sql.SQLException when a database related error occurs
	 * @exception java.io.IOException when an I/O error occurs
	 */
	public static void displayFile(Connection conn, String sqlTable, String sqlColumn,
		String sqlCondition, HttpServletResponse res) throws SQLException, IOException
	{
		PreparedStatement	stmt;
		ResultSet			rs;
		InputStream			in;
		OutputStream		out;
		String				val;
		byte[]				buf;
		int					nrRead;

		stmt = conn.prepareStatement("SELECT " + sqlColumn + "_ct, " + sqlColumn +
									" FROM " + sqlTable + " " + sqlCondition);
		try {
			stmt.setEscapeProcessing(false);
			rs = stmt.executeQuery();
			try {
				// set content-type
				val = rs.getString(1);
				if(val != null) {
					res.setContentType(val);
				} else {
					res.setContentType("text/plain");
				}

				// copy file data
				in = rs.getBinaryStream(2);
				try {
					out = res.getOutputStream();
					try {
						buf = new byte[8192];
						while((nrRead = in.read(buf, 0, buf.length)) != -1) {
							out.write(buf, 0, nrRead);
						}
					} finally {
						out.close();
					}
				} finally {
					in.close();
				}
			} finally {
				rs.close();
			}
		} finally {
			stmt.close();
		}
	}
}

