/*
	GnuJSP Toolkit - useful classes for JSP and servlet development
	Copyright (C) 1998, Vincent Partington <vinny@nmg.nl>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	nl.nmg.sql;

import	java.sql.*;
import	java.util.Properties;

/**
 * The PoolDriver handles JDBC URLs that start with "jdbc:pool:"
 * and are followed by another JDBC URL, e.g.
 * <CODE>jdbc:pool:jdbc:odbc:SampleTable</CODE>
 * and calls PoolDriverManager.getConnection() to get a pool
 * connection for the encapsulated JDBC URL.
 *
 * <P>For example, the following code samples do the same.
 *
 * <PRE>
 * DriverManager.registerDriver(new PoolDriver());
 * conn = DriverManager.getConnection("jdbc:pool:jdbc:odbc:SampleTable");
 * </PRE>
 * and
 * <PRE>
 * conn = PoolDriverManager.getConnection("jdbc:odbc:SampleTable");
 * </PRE>
 *
 * @see nl.nmg.sql.PoolDriverManager
 * @see java.sql.Driver
 */
public class PoolDriver implements Driver
{
	// Register driver with JDBC
	static
	{
		try {
			PoolDriver d = new PoolDriver();
			java.sql.DriverManager.registerDriver(d);
		} catch(SQLException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Creates an instance of the pool driver. The static initializer
	 * takes care of registering the driver with the driver manager.
	 *
	 * @exception SQLException never
	 */
	public PoolDriver() throws SQLException
	{
		// Driver has already been registered by static initializer
	}

	/**
	 * Return true if the PoolDriver can do something with this.
	 *
	 * @param url	the URL of the database to connect to
	 * @exception SQLException never
	 */
	public boolean acceptsURL(String url) throws SQLException
	{
		return url.startsWith("jdbc:pool:");
	}

	/**
	 * Open or reuse a proxy connection to the specified URL.
	 *
	 * @param url the URL of the database to connect to
	 * @param info connection properties
	 * @exception SQLException when PoolDriverManager.getConnection() throws
	 *			an exception
	 */
	public Connection connect(String url, Properties info) throws SQLException
	{
		String				realURL;

		// check whether we accept this URL
		if(!acceptsURL(url))
			return null;

		// get connection from pool
		return PoolDriverManager.getConnection(url.substring(10), info);
	}

	/**
	 * The getPropertyInfo method is intended to allow a generic GUI
	 * tool to discover what properties it should prompt a human for
	 * in order to get enough information to connect to a database.
	 *
	 * <p>Note that depending on the values the human has supplied so
	 * far, additional values may become necessary, so it may be necessary
	 * to iterate through several calls to getPropertyInfo
	 *
	 * @param url the Url of the database to connect to
	 * @param info a proposed list of tag/value pairs that will be sent on
	 * 	connect open.
	 * @return An array of DriverPropertyInfo objects describing
	 * 	possible properties.  This array may be an empty array if
	 *	no properties are required
	 * @exception SQLException never
	 * @see java.sql.Driver#getPropertyInfo
	 */
	public DriverPropertyInfo[] getPropertyInfo(String url, Properties info) throws SQLException
	{
		return new DriverPropertyInfo[0];
	}

	/**
	 * Gets the drivers major version number
	 *
	 * @return the drivers major version number
	 */
	public int getMajorVersion()
	{
		return PoolDriverManager.MAJOR_VERSION;
	}

	/**
	 * Get the drivers minor version number
	 *
	 * @return the drivers minor version number
	 */
	public int getMinorVersion()
	{
		return PoolDriverManager.MINOR_VERSION;
	}

	/**
	 * Report whether the driver is a genuine JDBC compliant driver.  A
	 * driver may only report "true" here if it passes the JDBC compliance
	 * tests, otherwise it is required to return false.  JDBC compliance
	 * requires full support for the JDBC API and full support for SQL 92
	 * Entry Level.
	 *
	 * <P>The PoolDriver does not know if the real driver used is JDBC
	 * compliant, the safest answer is false.
	 */
	public boolean jdbcCompliant()
	{
		return false;
	}
}
